package ru.dbotthepony.kstarbound.defs.world

import com.google.gson.Gson
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.gson.consumeNull
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.gson.value
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.collect.WeightedList
import ru.dbotthepony.kstarbound.json.builder.DispatchingAdapter
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import kotlin.properties.Delegates

enum class BeamUpRule(override val jsonName: String) : IStringSerializable {
	NOWHERE("Nowhere"),
	SURFACE("Surface"),
	ANYWHERE("Anywhere"),
	ANYWHERE_WITH_WARNING("AnywhereWithWarning");
}

enum class WorldEdgeForceRegion(override val jsonName: String) : IStringSerializable {
	NONE("None"),
	TOP("Top"),
	BOTTOM("Bottom"),
	TOP_AND_BOTTOM("TopAndBottom");
}

enum class VisitableWorldParametersType(override val jsonName: String, val token: TypeToken<out VisitableWorldParameters>) : IStringSerializable {
	TERRESTRIAL("TerrestrialWorldParameters", TypeToken.get(TerrestrialWorldParameters::class.java)),
	ASTEROIDS("AsteroidsWorldParameters", TypeToken.get(AsteroidsWorldParameters::class.java)),
	FLOATING_DUNGEON("FloatingDungeonWorldParameters", TypeToken.get(FloatingDungeonWorldParameters::class.java));

	companion object : TypeAdapterFactory {
		val ADAPTER = DispatchingAdapter("type", { type }, { token }, entries)

		override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (VisitableWorldParameters::class.java.isAssignableFrom(type.rawType)) {
				return object : TypeAdapter<VisitableWorldParameters>() {
					private val objects = gson.getAdapter(JsonObject::class.java)

					override fun write(out: JsonWriter, value: VisitableWorldParameters?) {
						if (value == null)
							out.nullValue()
						else
							out.value(value.toJson())
					}

					override fun read(`in`: JsonReader): VisitableWorldParameters? {
						if (`in`.consumeNull())
							return null

						val instance = type.rawType.getDeclaredConstructor().newInstance() as VisitableWorldParameters
						instance.fromJson(objects.read(`in`))
						return instance
					}
				} as TypeAdapter<T>
			}

			return null
		}
	}
}

// using notnull delegate because this will look very ugly
// if done as immutable class
abstract class VisitableWorldParameters {
	var threatLevel: Double = 0.0
		protected set
	var typeName: String by Delegates.notNull()
		protected set
	var worldSize: Vector2i by Delegates.notNull()
		protected set
	var gravity: Vector2d = Vector2d.ZERO
		protected set
	var airless: Boolean = false
		protected set
	var environmentStatusEffects: Set<String> by Delegates.notNull()
		protected set
	var overrideTech: Set<String>? = null
		protected set
	var globalDirectives: Set<String>? = null
		protected set
	var beamUpRule: BeamUpRule by Delegates.notNull()
		protected set
	var disableDeathDrops: Boolean = false
		protected set
	var terraformed: Boolean = false
		protected set
	var worldEdgeForceRegions: WorldEdgeForceRegion = WorldEdgeForceRegion.NONE
		protected set
	var weatherPool: WeightedList<String>? = null
		protected set

	abstract fun createLayout(seed: Long): WorldLayout

	abstract val type: VisitableWorldParametersType

	// lazy but effective, and less error prone
	@JsonFactory
	data class StoreData(
		val threatLevel: Double,
		val typeName: String,
		val worldSize: Vector2i,
		val gravity: Either<Double, Vector2d>,
		val airless: Boolean,
		val environmentStatusEffects: Set<String>,
		val overrideTech: Set<String>?,
		val globalDirectives: Set<String>?,
		val beamUpRule: BeamUpRule,
		val disableDeathDrops: Boolean,
		val terraformed: Boolean,
		val worldEdgeForceRegions: WorldEdgeForceRegion,
		val weatherPool: WeightedList<String>?,
	)

	open fun fromJson(data: JsonObject) {
		val read = Starbound.gson.fromJson(data, StoreData::class.java)

		this.threatLevel = read.threatLevel
		this.typeName = read.typeName
		this.worldSize = read.worldSize
		this.gravity = read.gravity.map({ Vector2d(y = it) }, { it })
		this.airless = read.airless
		this.environmentStatusEffects = read.environmentStatusEffects
		this.overrideTech = read.overrideTech
		this.globalDirectives = read.globalDirectives
		this.beamUpRule = read.beamUpRule
		this.disableDeathDrops = read.disableDeathDrops
		this.terraformed = read.terraformed
		this.worldEdgeForceRegions = read.worldEdgeForceRegions
		this.weatherPool = read.weatherPool
	}

	open fun toJson(data: JsonObject, isLegacy: Boolean = Starbound.IS_WRITING_LEGACY_JSON) {
		val store = StoreData(
			threatLevel,
			typeName,
			worldSize,
			if (isLegacy) Either.left(gravity.y) else Either.right(gravity),
			airless,
			environmentStatusEffects,
			overrideTech,
			globalDirectives,
			beamUpRule,
			disableDeathDrops,
			terraformed,
			worldEdgeForceRegions,
			weatherPool,
		)

		for ((k, v) in (Starbound.gson.toJsonTree(store) as JsonObject).entrySet()) {
			data[k] = v
		}

		data["type"] = type.jsonName
	}

	fun toJson(isLegacy: Boolean = Starbound.IS_WRITING_LEGACY_JSON): JsonObject {
		val data = JsonObject()
		toJson(data, isLegacy)
		return data
	}
}
